# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing data exxtractor base objects"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package base_fempostextractors
#  \ingroup FEM
#  \brief base objects for data extractors

from vtkmodules.vtkCommonCore import vtkIntArray
from vtkmodules.vtkCommonCore import vtkDoubleArray
from vtkmodules.vtkCommonDataModel import vtkTable

from PySide.QtCore import QT_TRANSLATE_NOOP

from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper

# helper functions
# ################


def is_extractor_object(obj):
    if not hasattr(obj, "Proxy"):
        return False

    return hasattr(obj.Proxy, "ExtractionType")


def get_extraction_type(obj):
    # returns the extractor type string, or throws exception if
    # not a extractor
    return obj.Proxy.ExtractionType


def get_extraction_dimension(obj):
    # returns the extractor dimension string, or throws exception if
    # not a extractor
    return obj.Proxy.ExtractionDimension


# Base class for all extractors with common source and table handling functionality
# Note: Never use directly, always subclass! This class does not create a
#       ExtractionType/Dimension variable, hence will not work correctly.
class Extractor(base_fempythonobject.BaseFemPythonObject):

    def __init__(self, obj):
        super().__init__(obj)
        self._setup_properties(obj)

    def _setup_properties(self, obj):
        pl = obj.PropertiesList
        for prop in self._get_properties():
            if not prop.name in pl:
                prop.add_to_object(obj)

    def _get_properties(self):

        prop = [
            _PropHelper(
                type="Fem::PropertyPostDataObject",
                name="Table",
                group="Base",
                doc=QT_TRANSLATE_NOOP("FEM", "The data table that stores the extracted data"),
                value=vtkTable(),
            ),
            _PropHelper(
                type="App::PropertyLink",
                name="Source",
                group="Base",
                doc=QT_TRANSLATE_NOOP("FEM", "The data source from which the data is extracted"),
                value=None,
            ),
        ]
        return prop

    def onDocumentRestored(self, obj):
        self._setup_properties(obj)

    def onChanged(self, obj, prop):

        if prop == "Source":
            # check if the source is a Post object
            if obj.Source and not obj.Source.isDerivedFrom("Fem::FemPostObject"):
                FreeCAD.Console.PrintWarning("Invalid object: Line source must be FemPostObject")
                obj.Source = None

    def get_vtk_table(self, obj):
        if not obj.DataTable:
            obj.DataTable = vtkTable()

        return obj.DataTable

    def component_options(self, num):

        match num:
            case 2:
                return ["X", "Y"]
            case 3:
                return ["X", "Y", "Z"]
            case 6:
                return ["XX", "YY", "ZZ", "XY", "XZ", "YZ"]
            case _:
                return ["Not a vector"]

    def get_representive_fieldname(self, obj):
        # should return the representative field name, e.g. Position (X)
        return ""


class Extractor1D(Extractor):

    ExtractionDimension = "1D"

    def __init__(self, obj):
        super().__init__(obj)

    def _get_properties(self):
        prop = [
            _PropHelper(
                type="App::PropertyEnumeration",
                name="XField",
                group="X Data",
                doc=QT_TRANSLATE_NOOP("FEM", "The field to use as X data"),
                value=[],
            ),
            _PropHelper(
                type="App::PropertyEnumeration",
                name="XComponent",
                group="X Data",
                doc=QT_TRANSLATE_NOOP(
                    "FEM", "Which part of the X field vector to use for the X axis"
                ),
                value=[],
            ),
        ]

        return super()._get_properties() + prop

    def onChanged(self, obj, prop):

        super().onChanged(obj, prop)

        if prop == "XField" and obj.Source and obj.Source.getDataSet():
            point_data = obj.Source.getDataSet().GetPointData()
            self._setup_x_component_property(obj, point_data)

        if prop == "Source":
            if obj.Source:
                dset = obj.Source.getDataSet()
                if dset:
                    self._setup_x_properties(obj, dset)
                else:
                    self._clear_x_properties(obj)
            else:
                self._clear_x_properties(obj)

    def _setup_x_component_property(self, obj, point_data):

        if obj.XField == "Index":
            obj.XComponent = self.component_options(1)
        elif obj.XField == "Position":
            obj.XComponent = self.component_options(3)
        else:
            array = point_data.GetAbstractArray(obj.XField)
            obj.XComponent = self.component_options(array.GetNumberOfComponents())

    def _clear_x_properties(self, obj):
        if hasattr(obj, "XComponent"):
            obj.XComponent = []
        if hasattr(obj, "XField"):
            obj.XField = []

    def _setup_x_properties(self, obj, dataset):
        # Set all X Data properties correctly for the given dataset
        fields = ["Index", "Position"]
        point_data = dataset.GetPointData()

        for i in range(point_data.GetNumberOfArrays()):
            fields.append(point_data.GetArrayName(i))

        current_field = obj.XField
        obj.XField = fields
        if current_field in fields:
            obj.XField = current_field

        self._setup_x_component_property(obj, point_data)

    def _x_array_component_to_table(self, obj, array, table):
        # extracts the component out of the array according to XComponent setting
        # Note: Uses the array name unchanged

        if array.GetNumberOfComponents() == 1:
            table.AddColumn(array)
        else:
            component_array = vtkDoubleArray()
            component_array.SetNumberOfComponents(1)
            component_array.SetNumberOfTuples(array.GetNumberOfTuples())
            c_idx = obj.getEnumerationsOfProperty("XComponent").index(obj.XComponent)
            component_array.CopyComponent(0, array, c_idx)
            component_array.SetName(array.GetName())
            table.AddColumn(component_array)

    def _x_array_from_dataset(self, obj, dataset, copy=True):
        # extracts the relevant array from the dataset and returns a copy
        # indices = None uses all indices, otherwise the values in this list

        match obj.XField:
            case "Index":
                # index needs always to be build, ignore copy argument
                num = dataset.GetPoints().GetNumberOfPoints()
                array = vtkIntArray()
                array.SetNumberOfTuples(num)
                array.SetNumberOfComponents(1)
                for i in range(num):
                    array.SetValue(i, i)

            case "Position":

                orig_array = dataset.GetPoints().GetData()
                if copy:
                    array = vtkDoubleArray()
                    array.DeepCopy(orig_array)
                else:
                    array = orig_array

            case _:
                point_data = dataset.GetPointData()
                orig_array = point_data.GetAbstractArray(obj.XField)
                if copy:
                    array = vtkDoubleArray()
                    array.DeepCopy(orig_array)
                else:
                    array = orig_array

        return array

    def get_representive_fieldname(self, obj):
        # representative field is the x field
        label = obj.XField
        if not label:
            return ""

        if len(obj.getEnumerationsOfProperty("XComponent")) > 1:
            label += f" ({obj.XComponent})"

        return label


class Extractor2D(Extractor1D):

    ExtractionDimension = "2D"

    def __init__(self, obj):
        super().__init__(obj)

    def _get_properties(self):
        prop = [
            _PropHelper(
                type="App::PropertyEnumeration",
                name="YField",
                group="Y Data",
                doc=QT_TRANSLATE_NOOP("FEM", "The field to use as Y data"),
                value=[],
            ),
            _PropHelper(
                type="App::PropertyEnumeration",
                name="YComponent",
                group="Y Data",
                doc=QT_TRANSLATE_NOOP(
                    "FEM", "Which part of the Y field vector to use for the Y axis"
                ),
                value=[],
            ),
        ]

        return super()._get_properties() + prop

    def onChanged(self, obj, prop):

        super().onChanged(obj, prop)

        if prop == "YField" and obj.Source and obj.Source.getDataSet():
            point_data = obj.Source.getDataSet().GetPointData()
            self._setup_y_component_property(obj, point_data)

        if prop == "Source":
            if obj.Source:
                dset = obj.Source.getDataSet()
                if dset:
                    self._setup_y_properties(obj, dset)
                else:
                    self._clear_y_properties(obj)
            else:
                self._clear_y_properties(obj)

    def _setup_y_component_property(self, obj, point_data):

        if obj.YField == "Position":
            obj.YComponent = self.component_options(3)
        else:
            array = point_data.GetAbstractArray(obj.YField)
            obj.YComponent = self.component_options(array.GetNumberOfComponents())

    def _clear_y_properties(self, obj):
        if hasattr(obj, "YComponent"):
            obj.YComponent = []
        if hasattr(obj, "YField"):
            obj.YField = []

    def _setup_y_properties(self, obj, dataset):
        # Set all X Data properties correctly for the given dataset
        fields = ["Position"]
        point_data = dataset.GetPointData()

        for i in range(point_data.GetNumberOfArrays()):
            fields.append(point_data.GetArrayName(i))

        current_field = obj.YField
        obj.YField = fields
        if current_field in fields:
            obj.YField = current_field

        self._setup_y_component_property(obj, point_data)

    def _y_array_component_to_table(self, obj, array, table):
        # extracts the component out of the array according to XComponent setting

        if array.GetNumberOfComponents() == 1:
            table.AddColumn(array)
        else:
            component_array = vtkDoubleArray()
            component_array.SetNumberOfComponents(1)
            component_array.SetNumberOfTuples(array.GetNumberOfTuples())
            c_idx = obj.getEnumerationsOfProperty("YComponent").index(obj.YComponent)
            component_array.CopyComponent(0, array, c_idx)
            component_array.SetName(array.GetName())
            table.AddColumn(component_array)

    def _y_array_from_dataset(self, obj, dataset, copy=True):
        # extracts the relevant array from the dataset and returns a copy
        # indices = None uses all indices, otherwise the values in this list

        match obj.YField:
            case "Position":

                orig_array = dataset.GetPoints().GetData()
                if copy:
                    array = vtkDoubleArray()
                    array.DeepCopy(orig_array)
                else:
                    array = orig_array

            case _:
                point_data = dataset.GetPointData()
                orig_array = point_data.GetAbstractArray(obj.YField)

                if copy:
                    array = vtkDoubleArray()
                    array.DeepCopy(orig_array)
                else:
                    array = orig_array

        return array

    def get_representive_fieldname(self, obj):
        # representative field is the y field
        label = obj.YField
        if not label:
            return ""

        if len(obj.getEnumerationsOfProperty("YComponent")) > 1:
            label += f" ({obj.YComponent})"

        return label



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing data visualization base object"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package base_fempostextractors
#  \ingroup FEM
#  \brief base objects for data visualizations

from vtkmodules.vtkCommonDataModel import vtkTable
from vtkmodules.vtkCommonCore import vtkDoubleArray

from . import base_fempythonobject
from . import base_fempostextractors

# helper functions
# ################


def is_visualization_object(obj):
    if not obj:
        return False

    if not hasattr(obj, "Proxy"):
        return False

    return hasattr(obj.Proxy, "VisualizationType")


def get_visualization_type(obj):
    # returns the extractor type string, or throws exception if
    # not a extractor
    return obj.Proxy.VisualizationType


def is_visualization_extractor_type(obj, vistype):

    # must be extractor
    if not base_fempostextractors.is_extractor_object(obj):
        return False

    # must be visualization object
    if not is_visualization_object(obj):
        return False

    # must be correct type
    if get_visualization_type(obj) != vistype:
        return False

    return True


# Base class for all visualizations
# It collects all data from its extraction objects into a table.
# Note: Never use directly, always subclass! This class does not create a
#       Visualization variable, hence will not work correctly.
class PostVisualization(base_fempythonobject.BaseFemPythonObject):

    def __init__(self, obj):
        super().__init__(obj)
        self.Type = "Fem::FemPostVisualization"
        obj.addExtension("App::GroupExtensionPython")
        self._setup_properties(obj)

    def _setup_properties(self, obj):
        pl = obj.PropertiesList
        for prop in self._get_properties():
            if not prop.name in pl:
                prop.add_to_object(obj)

    def _get_properties(self):
        # override if subclass wants to add additional properties

        prop = [
            base_fempostextractors._PropHelper(
                type="Fem::PropertyPostDataObject",
                name="Table",
                group="Base",
                doc="The data table that stores the data for visualization",
                value=vtkTable(),
            ),
        ]
        return prop

    def onDocumentRestored(self, obj):
        # if a new property was added we handle it by setup
        # Override if subclass needs to handle changed property type

        self._setup_properties(obj)

    def onChanged(self, obj, prop):
        # Ensure only correct child object types are in the group

        if prop == "Group":
            # check if all objects are allowed

            children = obj.Group
            for child in obj.Group:
                if not is_visualization_extractor_type(child, self.VisualizationType):
                    FreeCAD.Console.PrintWarning(
                        f"{child.Label} is not a {self.VisualizationType} extraction object, cannot be added"
                    )
                    children.remove(child)

            if len(obj.Group) != len(children):
                obj.Group = children

    def execute(self, obj):
        # Collect all extractor child data into our table
        # Note: Each childs table can have different number of rows. We need
        # to pad the date for our table in this case

        rows = self.getLongestColumnLength(obj)
        table = vtkTable()
        for child in obj.Group:

            # If child has no Source, its table should be empty. However,
            # it would theoretical be possible that child source was set
            # to none without recompute, and the visualization was manually
            # recomputed afterwards
            if not child.Source and (child.Table.GetNumberOfColumns() > 0):
                FreeCAD.Console.PrintWarning(
                    f"{child.Label} has data, but no Source object. Will be ignored"
                )
                continue

            c_table = child.Table
            for i in range(c_table.GetNumberOfColumns()):
                c_array = c_table.GetColumn(i)
                array = vtkDoubleArray()

                if c_array.GetNumberOfTuples() == rows:
                    # simple deep copy is enough
                    array.DeepCopy(c_array)

                else:
                    array.SetNumberOfComponents(c_array.GetNumberOfComponents())
                    array.SetNumberOfTuples(rows)
                    array.Fill(0)  # so that all non-used entries are set to 0
                    for j in range(c_array.GetNumberOfTuples()):
                        array.SetTuple(j, c_array.GetTuple(j))

                array.SetName(f"{child.Source.Name}: {c_array.GetName()}")
                table.AddColumn(array)

        obj.Table = table
        return False

    def getLongestColumnLength(self, obj):
        # iterate all extractor children and get the column lengths

        length = 0
        for child in obj.Group:
            if base_fempostextractors.is_extractor_object(child):
                table = child.Table
                if table.GetNumberOfColumns() > 0:
                    # we assume all columns of an extractor have same length
                    num = table.GetColumn(0).GetNumberOfTuples()
                    if num > length:
                        length = num

        return length



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix write inpfile footer"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


import os
import time


def write_footer(f, ccxwriter):

    f.write("\n{}\n".format(59 * "*"))
    f.write("** CalculiX Input file\n")
    f.write(
        "**   written by    --> FreeCAD {}.{}.{}\n".format(
            ccxwriter.fc_ver[0], ccxwriter.fc_ver[1], ccxwriter.fc_ver[2]
        )
    )
    f.write(f"**   written on    --> {time.ctime()}\n")
    f.write(f"**   file name     --> {os.path.basename(ccxwriter.document.FileName)}\n")
    f.write(f"**   analysis name --> {ccxwriter.analysis.Name}\n")
    f.write("**\n")
    f.write("**\n")
    f.write(ccxwriter.units_information)
    f.write("**\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Door command"""

import FreeCAD
import FreeCADGui
from bimcommands import BimWindow

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_Door(BimWindow.Arch_Window):

    def __init__(self):
        super().__init__()
        self.doormode = True

    def GetResources(self):
        return {
            "Pixmap": "BIM_Door",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Door", "Door"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_Door", "Places a door at a given location"),
            "Accel": "D,O",
        }


FreeCADGui.addCommand("BIM_Door", BIM_Door())



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD post glyph filter"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package post_glyphfilter
#  \ingroup FEM
#  \brief Post processing filter creating glyphs for vector fields

import FreeCAD

# check vtk version to potentially find missmatchs
from femguiutils.vtk_module_handling import vtk_module_handling

vtk_module_handling()

# IMPORTANT: Never import vtk directly. Often vtk is compiled with different QT
# version than FreeCAD, and "import vtk" crashes by importing qt components.
# Always import the filter and data modules only.
from vtkmodules.vtkFiltersCore import vtkMaskPoints
from vtkmodules.vtkFiltersCore import vtkGlyph3D
import vtkmodules.vtkFiltersSources as vtkSources

from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class PostGlyphFilter(base_fempythonobject.BaseFemPythonObject):
    """
    A post processing filter adding glyphs
    """

    Type = "Fem::PostFilterPython"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

        self.__setupFilterPipeline(obj)

    def _get_properties(self):

        prop = [
            _PropHelper(
                type="App::PropertyEnumeration",
                name="Glyph",
                group="Glyph",
                doc="The form of the glyph",
                value=["Arrow", "Cone", "Cube", "Cylinder", "Line", "Sphere"],
            ),
            _PropHelper(
                type="App::PropertyEnumeration",
                name="OrientationData",
                group="Glyph",
                doc="Which vector field is used to orient the glyphs",
                value=["None"],
            ),
            _PropHelper(
                type="App::PropertyEnumeration",
                name="ScaleData",
                group="Scale",
                doc="Which data field is used to scale the glyphs",
                value=["None"],
            ),
            _PropHelper(
                type="App::PropertyEnumeration",
                name="VectorScaleMode",
                group="Scale",
                doc="If the scale data is a vector, this property decides if the glyph is scaled by vector magnitude or by the individual components",
                value=["Not a vector"],
            ),
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="ScaleFactor",
                group="Scale",
                doc="A constant multiplier the glyphs are scaled with",
                value=(1, 0, 1e12, 1e-12),
            ),
            _PropHelper(
                type="App::PropertyEnumeration",
                name="MaskMode",
                group="Masking",
                doc="Which vertices are used as glyph locations",
                value=["Use All", "Every Nth", "Uniform Sampling"],
            ),
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="Stride",
                group="Masking",
                doc='Define the stride for "Every Nth" masking mode',
                value=(2, 1, 999999999, 1),
            ),
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="MaxNumber",
                group="Masking",
                doc='Defines the maximum number of vertices used for "Uniform Sampling" masking mode',
                value=(1000, 1, 999999999, 1),
            ),
        ]
        return prop

    def __setupMaskingFilter(self, obj, masking):

        if obj.MaskMode == "Use All":
            masking.RandomModeOff()
            masking.SetOnRatio(1)
            masking.SetMaximumNumberOfPoints(int(1e10))
        elif obj.MaskMode == "Every Nth":
            masking.RandomModeOff()
            masking.SetOnRatio(obj.Stride)
            masking.SetMaximumNumberOfPoints(int(1e10))
        else:
            masking.SetOnRatio(1)
            masking.SetMaximumNumberOfPoints(obj.MaxNumber)
            masking.RandomModeOn()

    def __setupGlyphFilter(self, obj, glyph):

        # scaling
        if obj.ScaleData != "None":

            glyph.ScalingOn()
            if obj.ScaleData in obj.getInputVectorFields():

                # make sure the vector mode is set correctly
                if obj.VectorScaleMode == "Not a vector":
                    obj.VectorScaleMode = ["Scale by magnitude", "Scale by components"]
                    obj.VectorScaleMode = "Scale by magnitude"

                if obj.VectorScaleMode == "Scale by magnitude":
                    glyph.SetScaleModeToScaleByVector()
                else:
                    glyph.SetScaleModeToScaleByVectorComponents()

                glyph.SetInputArrayToProcess(2, 0, 0, 0, obj.ScaleData)

            else:
                # scalar scaling mode
                if obj.VectorScaleMode != "Not a vector":
                    obj.VectorScaleMode = ["Not a vector"]

                glyph.SetInputArrayToProcess(2, 0, 0, 0, obj.ScaleData)
                glyph.SetScaleModeToScaleByScalar()
        else:
            glyph.ScalingOff()

        glyph.SetScaleFactor(obj.ScaleFactor)

        # Orientation
        if obj.OrientationData != "None":
            glyph.OrientOn()
            glyph.SetInputArrayToProcess(1, 0, 0, 0, obj.OrientationData)
        else:
            glyph.OrientOff()

    def __setupFilterPipeline(self, obj):

        # store of all algorithms for later access
        # its map filter_name : [source, mask, glyph]
        self._algorithms = {}

        # create all vtkalgorithm combinations and set them as filter pipeline
        sources = {
            "Arrow": vtkSources.vtkArrowSource,
            "Cone": vtkSources.vtkConeSource,
            "Cube": vtkSources.vtkCubeSource,
            "Cylinder": vtkSources.vtkCylinderSource,
            "Line": vtkSources.vtkLineSource,
            "Sphere": vtkSources.vtkSphereSource,
        }

        for source_name in sources:

            source = sources[source_name]()

            masking = vtkMaskPoints()
            self.__setupMaskingFilter(obj, masking)

            glyph = vtkGlyph3D()
            glyph.SetSourceConnection(source.GetOutputPort(0))
            glyph.SetInputConnection(masking.GetOutputPort(0))
            self.__setupGlyphFilter(obj, glyph)

            self._algorithms[source_name] = [source, masking, glyph]
            obj.addFilterPipeline(source_name, masking, glyph)

        obj.setActiveFilterPipeline(obj.Glyph)

    def onDocumentRestored(self, obj):
        # resetup the pipeline
        self.__setupFilterPipeline(obj)

    def execute(self, obj):
        # we check what new inputs

        vector_fields = obj.getInputVectorFields()
        all_fields = vector_fields + obj.getInputScalarFields()

        vector_fields.sort()
        all_fields.sort()

        current_orient = obj.OrientationData
        enumeration = ["None"] + vector_fields
        obj.OrientationData = enumeration
        if current_orient in enumeration:
            obj.OrientationData = current_orient

        current_scale = obj.ScaleData
        enumeration = ["None"] + all_fields
        obj.ScaleData = enumeration
        if current_scale in enumeration:
            obj.ScaleData = current_scale

        # make sure parent class execute is called!
        return False

    def onChanged(self, obj, prop):

        # check if we are setup already
        if not hasattr(self, "_algorithms"):
            return

        if prop == "Glyph":
            obj.setActiveFilterPipeline(obj.Glyph)

        if prop == "MaskMode":
            for filter in self._algorithms:
                masking = self._algorithms[filter][1]
                self.__setupMaskingFilter(obj, masking)

        if prop == "Stride":
            # if mode is use all stride setting needs to stay at one
            if obj.MaskMode == "Every Nth":
                for filter in self._algorithms:
                    masking = self._algorithms[filter][1]
                    masking.SetOnRatio(obj.Stride)

        if prop == "MaxNumber":
            if obj.MaskMode == "Uniform Sampling":
                for filter in self._algorithms:
                    masking = self._algorithms[filter][1]
                    masking.SetMaximumNumberOfPoints(obj.MaxNumber)

        if prop == "OrientationData" or prop == "ScaleData":
            for filter in self._algorithms:
                glyph = self._algorithms[filter][2]
                self.__setupGlyphFilter(obj, glyph)

        if prop == "ScaleFactor":
            for filter in self._algorithms:
                glyph = self._algorithms[filter][2]
                glyph.SetScaleFactor(obj.ScaleFactor)



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing view for summarizing extractor links"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package data_extraction
#  \ingroup FEM
#  \brief A widget that shows summaries of all available links to extractors

from PySide import QtGui, QtCore

import femobjects.base_fempostextractors as extr
import femobjects.base_fempostvisualizations as vis

import FreeCAD
import FreeCADGui

from . import post_visualization as pv

translate = FreeCAD.Qt.translate

# a model showing available visualizations and possible extractions
# #################################################################


def build_new_visualization_tree_model():
    # model that shows all options to create new visualizations

    model = QtGui.QStandardItemModel()

    visualizations = pv.get_registered_visualizations()
    for vis_name in visualizations:
        vis_icon = FreeCADGui.getIcon(visualizations[vis_name].icon)
        vis_item = QtGui.QStandardItem(vis_icon, translate("FEM", "New {}").format(vis_name))
        vis_item.setFlags(QtGui.Qt.ItemIsEnabled)
        vis_item.setData(visualizations[vis_name])

        for ext in visualizations[vis_name].extractions:
            icon = FreeCADGui.getIcon(ext.icon)
            name = ext.name.removeprefix(vis_name)
            ext_item = QtGui.QStandardItem(icon, translate("FEM", "with {}").format(name))
            ext_item.setFlags(QtGui.Qt.ItemIsEnabled)
            ext_item.setData(ext)
            vis_item.appendRow(ext_item)
        model.appendRow(vis_item)

    return model


def build_add_to_visualization_tree_model():
    # model that shows all possible visualization objects to add data to

    visualizations = pv.get_registered_visualizations()
    model = QtGui.QStandardItemModel()

    for obj in FreeCAD.ActiveDocument.Objects:
        if obj.isDerivedFrom("Fem::FemAnalysis"):
            ana_item = QtGui.QStandardItem(obj.ViewObject.Icon, obj.Label)
            ana_item.setFlags(QtGui.Qt.ItemIsEnabled)

            # check all children it it is a visualization
            for child in obj.Group:
                if vis.is_visualization_object(child):

                    vis_item = QtGui.QStandardItem(child.ViewObject.Icon, child.Label)
                    vis_type = vis.get_visualization_type(child)
                    vis_item.setFlags(QtGui.Qt.ItemIsEnabled)
                    vis_item.setData(child)
                    ana_item.appendRow(vis_item)

                    # add extractor items
                    for ext in visualizations[vis_type].extractions:
                        icon = FreeCADGui.getIcon(ext.icon)
                        name = ext.name.removeprefix(vis_type)
                        ext_item = QtGui.QStandardItem(
                            icon, translate("FEM", "Add {}").format(name)
                        )
                        ext_item.setFlags(QtGui.Qt.ItemIsEnabled)
                        ext_item.setData(ext)
                        vis_item.appendRow(ext_item)

            if ana_item.rowCount():
                model.appendRow(ana_item)

    return model


def build_post_object_item(post_object, extractions, vis_type):

    # definitely build a item and add the extractions
    post_item = QtGui.QStandardItem(
        post_object.ViewObject.Icon, translate("FEM", "From {}").format(post_object.Label)
    )
    post_item.setFlags(QtGui.Qt.ItemIsEnabled)
    post_item.setData(post_object)

    # add extractor items
    for ext in extractions:
        icon = FreeCADGui.getIcon(ext.icon)
        name = ext.name.removeprefix(vis_type)
        ext_item = QtGui.QStandardItem(icon, translate("FEM", "add {}").format(name))
        ext_item.setFlags(QtGui.Qt.ItemIsEnabled)
        ext_item.setData(ext)
        post_item.appendRow(ext_item)

    # if we are a post group, we need to add the children
    if post_object.hasExtension("Fem::FemPostGroupExtension"):

        for child in post_object.Group:
            if child.isDerivedFrom("Fem::FemPostObject"):
                item = build_post_object_item(child, extractions, vis_type)
                post_item.appendRow(item)

    return post_item


def build_add_from_data_tree_model(vis_type):
    # model that shows all Post data objects from which data can be extracted
    extractions = pv.get_registered_visualizations()[vis_type].extractions

    model = QtGui.QStandardItemModel()
    for obj in FreeCAD.ActiveDocument.Objects:
        if obj.isDerivedFrom("Fem::FemAnalysis"):
            ana_item = QtGui.QStandardItem(obj.ViewObject.Icon, obj.Label)
            ana_item.setFlags(QtGui.Qt.ItemIsEnabled)

            # check all children if it is a post object
            for child in obj.Group:
                if child.isDerivedFrom("Fem::FemPostObject"):
                    item = build_post_object_item(child, extractions, vis_type)
                    ana_item.appendRow(item)

            if ana_item.rowCount():
                model.appendRow(ana_item)

    return model


# implementation of GUI and its functionality
# ###########################################


class _ElideToolButton(QtGui.QToolButton):
    # tool button that elides its text, and left align icon and text

    def __init__(self, icon, text, parent):
        super().__init__(parent)

        self._text = text
        self._icon = icon

    def setCustomText(self, text):
        self._text = text
        self.repaint()

    def setCustomIcon(self, icon):
        self._icon = icon
        self.repaint()

    def sizeHint(self):
        button_size = super().sizeHint()
        icn_size = self.iconSize()
        min_margin = max((button_size - icn_size).height(), 6)
        return QtCore.QSize(self.iconSize().width() + 10, icn_size.height() + min_margin)

    def paintEvent(self, event):

        # draw notmal button, without text and icon
        super().paintEvent(event)

        # add icon and elided text
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
        painter.setRenderHint(QtGui.QPainter.SmoothPixmapTransform, True)

        margin = (self.height() - self.iconSize().height()) / 2
        icn_width = self.iconSize().width()
        if self._icon.isNull():
            icn_width = 0

        fm = self.fontMetrics()
        txt_size = self.width() - icn_width - 2 * margin
        if not self._icon.isNull():
            # we add the margin between icon and text
            txt_size -= margin

        txt_min = fm.boundingRect("…").width()

        # should we center the icon?
        xpos = margin
        if not self._icon.isNull() and txt_size < txt_min:
            # center icon
            xpos = self.width() / 2 - self.iconSize().width() / 2

        if not self._icon.isNull():
            match type(self._icon):
                case QtGui.QPixmap:
                    painter.drawPixmap(xpos, margin, self._icon.scaled(self.iconSize()))
                    xpos += self.iconSize().width()
                case QtGui.QIcon:
                    self._icon.paint(
                        painter, QtCore.QRect(QtCore.QPoint(margin, margin), self.iconSize())
                    )
                    xpos += self.iconSize().width()

            xpos += margin  # the margin to the text

        if txt_size >= txt_min:
            text = fm.elidedText(self._text, QtGui.Qt.ElideMiddle, txt_size)
            painter.drawText(xpos, margin + fm.ascent(), text)

        painter.end()


class _TreeChoiceButton(QtGui.QToolButton):

    selection = QtCore.Signal(object, object)

    def __init__(self, model):
        super().__init__()

        self.model = model
        self.setEnabled(bool(model.rowCount()))

        self.__skip_next_hide = False

        self.tree_view = QtGui.QTreeView(self)
        self.tree_view.setModel(model)

        self.tree_view.setFrameShape(QtGui.QFrame.NoFrame)
        self.tree_view.setHeaderHidden(True)
        self.tree_view.setSelectionBehavior(QtGui.QTreeView.SelectionBehavior.SelectRows)
        self.tree_view.expandAll()
        self.tree_view.clicked.connect(self.selectIndex)

        style = self.style()
        if not style.styleHint(QtGui.QStyle.SH_ItemView_ActivateItemOnSingleClick):
            self.tree_view.activated.connect(self.selectIndex)

        # set a complex menu
        self.popup = QtGui.QWidgetAction(self)
        self.popup.setDefaultWidget(self.tree_view)
        self.setPopupMode(QtGui.QToolButton.InstantPopup)
        self.addAction(self.popup)

    QtCore.Slot(QtCore.QModelIndex)

    def selectIndex(self, index):
        item = self.model.itemFromIndex(index)

        if item and not item.hasChildren():
            extraction = item.data()
            parent = item.parent().data()
            self.selection.emit(parent, extraction)
            self.popup.trigger()

    def setModel(self, model):
        self.model = model
        self.tree_view.setModel(model)
        self.tree_view.expandAll()

        # check if we should be disabled
        self.setEnabled(bool(model.rowCount()))


class _SettingsPopup(QtGui.QMenu):

    close = QtCore.Signal()

    def __init__(self, setting, parent):
        super().__init__(parent)

        self._setting = setting
        self.setWindowFlags(QtGui.Qt.Popup)
        self.setFocusPolicy(QtGui.Qt.ClickFocus)

        vbox = QtGui.QVBoxLayout()
        vbox.addWidget(setting)

        buttonBox = QtGui.QDialogButtonBox()
        buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Ok)
        buttonBox.accepted.connect(self.hide)
        vbox.addWidget(buttonBox)

        widget = QtGui.QFrame()
        widget.setLayout(vbox)

        vbox2 = QtGui.QVBoxLayout()
        vbox2.setContentsMargins(0, 0, 0, 0)
        vbox2.addWidget(widget)
        self.setLayout(vbox2)

    def size(self):
        return self._setting.sizeHint()

    def showEvent(self, event):
        # required to get keyboard events
        self.setFocus()

    def hideEvent(self, event):
        # emit on hide: this happens for OK button as well as
        # "click away" closing of the popup
        self.close.emit()

    def keyPressEvent(self, event):
        # close on hitting enter
        if event.key() == QtGui.Qt.Key_Enter or event.key() == QtGui.Qt.Key_Return:
            self.hide()


class _SummaryWidget(QtGui.QWidget):

    delete = QtCore.Signal(object, object)  # to delete: document object, summary widget

    def __init__(self, st_object, extractor, post_dialog):
        super().__init__()

        self._st_object = st_object
        self._extractor = extractor
        self._post_dialog = post_dialog

        extr_label = extractor.Proxy.get_representive_fieldname(extractor)
        extr_repr = extractor.ViewObject.Proxy.get_preview()

        # build the UI
        hbox = QtGui.QHBoxLayout()
        hbox.setContentsMargins(6, 0, 6, 0)
        hbox.setSpacing(2)

        self.extrButton = self._button(extractor.ViewObject.Icon, extr_label)
        self.viewButton = self._button(extr_repr[0], extr_repr[1], 1)

        size = self.viewButton.iconSize()
        size.setWidth(size.width() * 2)
        self.viewButton.setIconSize(size)

        if st_object:
            self.stButton = self._button(st_object.ViewObject.Icon, st_object.Label)
            hbox.addWidget(self.stButton)

        else:
            # that happens if the source of the extractor was deleted and now
            # that property is set to None
            self.extrButton.hide()
            self.viewButton.hide()

            self.warning = QtGui.QLabel(self)
            self.warning.full_text = translate("FEM", "{}: Data source not available").format(
                extractor.Label
            )
            hbox.addWidget(self.warning)

        self.rmButton = QtGui.QToolButton(self)
        self.rmButton.setIcon(FreeCADGui.getIcon("delete.svg"))
        self.rmButton.setAutoRaise(True)

        hbox.addWidget(self.extrButton)
        hbox.addWidget(self.viewButton)
        hbox.addSpacing(15)
        hbox.addWidget(self.rmButton)

        # add the separation line
        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setSpacing(5)
        vbox.addItem(hbox)
        self.frame = QtGui.QFrame(self)
        self.frame.setFrameShape(QtGui.QFrame.HLine)
        vbox.addWidget(self.frame)

        policy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)
        self.setSizePolicy(policy)
        # self.setMinimumSize(self.extrButton.sizeHint()+self.frame.sizeHint()*3)
        self.setLayout(vbox)

        # connect actions. We add functions to widget, as well as the data we need,
        # and use those as callback. This way every widget knows which objects to use
        if st_object:
            self.stButton.clicked.connect(self.showVisualization)
            self.extrButton.clicked.connect(self.editApp)
            self.viewButton.clicked.connect(self.editView)

        self.rmButton.clicked.connect(self.deleteTriggered)

        # make sure initial drawing happened
        # self._redraw()

    def _button(self, icon, text, stretch=2):

        btn = _ElideToolButton(icon, text, self)
        btn.setMinimumWidth(0)
        btn.setAutoRaise(True)
        btn.setToolTip(text)

        policy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)
        policy.setHorizontalStretch(stretch)
        btn.setSizePolicy(policy)
        return btn

    @QtCore.Slot()
    def showVisualization(self):
        if vis.is_visualization_object(self._st_object):
            # show the visualization
            self._st_object.ViewObject.Proxy.show_visualization()
        else:
            # for now just select the thing
            FreeCADGui.Selection.clearSelection()
            FreeCADGui.Selection.addSelection(self._st_object)

    def _position_dialog(self, dialog):

        # the scroll area does mess the mapping to global up, somehow
        # the transformation from the widget ot the scroll area gives
        # very weird values. Hence we build the coords of the widget
        # ourself

        summary = dialog.parent()  # == self
        base_widget = summary.parent()
        viewport = summary.parent()
        scroll = viewport.parent()

        top_left = (
            summary.geometry().topLeft()
            + base_widget.geometry().topLeft()
            + viewport.geometry().topLeft()
        )
        delta = (summary.width() - dialog.size().width()) / 2
        local_point = QtCore.QPoint(top_left.x() + delta, top_left.y() + summary.height())
        global_point = scroll.mapToGlobal(local_point)

        dialog.setGeometry(QtCore.QRect(global_point, dialog.sizeHint()))

    @QtCore.Slot()
    def editApp(self):
        if not hasattr(self, "appDialog"):
            widget = self._extractor.ViewObject.Proxy.get_app_edit_widget(self._post_dialog)
            self.appDialog = _SettingsPopup(widget, self)
            self.appDialog.close.connect(self.appAccept)

        if not self.appDialog.isVisible():
            # position correctly and show
            self._position_dialog(self.appDialog)
            self.appDialog.show()

    @QtCore.Slot()
    def editView(self):

        if not hasattr(self, "viewDialog"):
            widget = self._extractor.ViewObject.Proxy.get_view_edit_widget(self._post_dialog)
            self.viewDialog = _SettingsPopup(widget, self)
            self.viewDialog.close.connect(self.viewAccept)

        if not self.viewDialog.isVisible():
            # position correctly and show
            self._position_dialog(self.viewDialog)
            self.viewDialog.show()

    @QtCore.Slot()
    def deleteTriggered(self):
        self.delete.emit(self._extractor, self)

    @QtCore.Slot()
    def viewAccept(self):

        # update the preview
        extr_repr = self._extractor.ViewObject.Proxy.get_preview()
        self.viewButton.setCustomIcon(extr_repr[0])
        self.viewButton.setCustomText(extr_repr[1])
        self.viewButton.setToolTip(extr_repr[1])

    @QtCore.Slot()
    def appAccept(self):

        # update the preview
        extr_label = self._extractor.Proxy.get_representive_fieldname(self._extractor)
        self.extrButton.setCustomText(extr_label)
        self.extrButton.setToolTip(extr_label)


class ExtractLinkView(QtGui.QWidget):

    def __init__(self, obj, is_source, post_dialog):
        # initializes the view.
        # obj: The object for which the links should be shown / summarized
        # is_source: Bool, if the object is the data source (e.g. postobject), or the target (e.g. plots)

        super().__init__()

        self._object = obj
        self._is_source = is_source
        self._post_dialog = post_dialog
        self._widgets = []

        # build the layout:
        self._scroll_view = QtGui.QScrollArea(self)
        self._scroll_view.setHorizontalScrollBarPolicy(QtGui.Qt.ScrollBarAlwaysOff)
        self._scroll_view.setWidgetResizable(True)
        self._scroll_widget = QtGui.QWidget(self._scroll_view)
        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(0, 6, 0, 0)
        vbox.addStretch()
        self._scroll_widget.setLayout(vbox)
        self._scroll_view.setWidget(self._scroll_widget)

        hbox = QtGui.QHBoxLayout()
        hbox.setSpacing(6)
        label = QtGui.QLabel(translate("FEM", "Data used in:"))
        if not self._is_source:
            label.setText(translate("FEM", "Data used from:"))

        label.setAlignment(QtGui.Qt.AlignBottom)
        hbox.addWidget(label)
        hbox.addStretch()

        if self._is_source:

            self._add = _TreeChoiceButton(build_add_to_visualization_tree_model())
            self._add.setText(translate("FEM", "Add data to"))
            self._add.selection.connect(self.addExtractionToVisualization)
            hbox.addWidget(self._add)

            self._create = _TreeChoiceButton(build_new_visualization_tree_model())
            self._create.setText(translate("FEM", "New"))
            self._create.selection.connect(self.newVisualization)
            hbox.addWidget(self._create)

        else:
            vis_type = vis.get_visualization_type(self._object)
            self._add = _TreeChoiceButton(build_add_from_data_tree_model(vis_type))
            self._add.setText(translate("FEM", "Add data from"))
            self._add.selection.connect(self.addExtractionToPostObject)
            hbox.addWidget(self._add)

        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addItem(hbox)
        vbox.addWidget(self._scroll_view)

        self.setLayout(vbox)

        # add the content
        self.repopulate()

    def _build_summary_widget(self, extractor):

        if self._is_source:
            st_object = extractor.getParentGroup()
        else:
            st_object = extractor.Source

        widget = _SummaryWidget(st_object, extractor, self._post_dialog)
        widget.delete.connect(self._delete_extraction)

        return widget

    def _delete_extraction(self, extractor, widget):
        # remove the document object
        doc = extractor.Document
        doc.removeObject(extractor.Name)
        doc.recompute()

        # remove the widget
        self._widgets.remove(widget)
        widget.deleteLater()

    def repopulate(self):
        # collect all links that are available and shows them

        # clear the view
        for widget in self._widgets:
            widget.hide()
            widget.deleteLater()

        self._widgets = []

        # rebuild the widgets

        if self._is_source:
            candidates = self._object.InList
        else:
            candidates = self._object.OutList

        # get all widgets from the candidates
        for candidate in candidates:
            if extr.is_extractor_object(candidate):
                summary = self._build_summary_widget(candidate)
                self._widgets.append(summary)

        # fill the scroll area
        vbox = self._scroll_widget.layout()
        for widget in reversed(self._widgets):
            vbox.insertWidget(0, widget)

        # also reset the add button model
        if self._is_source:
            self._add.setModel(build_add_to_visualization_tree_model())

    def _find_parent_analysis(self, obj):
        # iterate upwards, till we find a analysis
        for parent in obj.InList:
            if parent.isDerivedFrom("Fem::FemAnalysis"):
                return parent

            analysis = self._find_parent_analysis(parent)
            if analysis:
                return analysis

        return None

    QtCore.Slot(object, object)  # visualization data, extraction data

    def newVisualization(self, vis_data, ext_data):

        FreeCADGui.addModule(vis_data.module)
        FreeCADGui.addModule(ext_data.module)
        FreeCADGui.addModule("FemGui")

        # create visualization
        FreeCADGui.doCommand(
            f"visualization = {vis_data.module}.{vis_data.factory}(FreeCAD.ActiveDocument)"
        )

        analysis = self._find_parent_analysis(self._object)
        if analysis:
            FreeCADGui.doCommand(f"FreeCAD.ActiveDocument.{analysis.Name}.addObject(visualization)")

        # create extraction and add it
        FreeCADGui.doCommand(
            f"extraction = {ext_data.module}.{ext_data.factory}(FreeCAD.ActiveDocument)"
        )
        FreeCADGui.doCommand(f"extraction.Source = FreeCAD.ActiveDocument.{self._object.Name}")
        # default values: color
        color_prop = FreeCADGui.ActiveDocument.ActiveObject.Proxy.get_default_color_property()
        if color_prop:
            FreeCADGui.doCommand(
                f"extraction.ViewObject.{color_prop} = visualization.ViewObject.Proxy.get_next_default_color()"
            )

        FreeCADGui.doCommand(f"visualization.addObject(extraction)")

        self._post_dialog._recompute()
        self.repopulate()

    QtCore.Slot(object, object)  # visualization object, extraction data

    def addExtractionToVisualization(self, vis_obj, ext_data):

        FreeCADGui.addModule(ext_data.module)
        FreeCADGui.addModule("FemGui")

        # create extraction and add it
        FreeCADGui.doCommand(
            f"extraction = {ext_data.module}.{ext_data.factory}(FreeCAD.ActiveDocument)"
        )
        FreeCADGui.doCommand(f"extraction.Source = FreeCAD.ActiveDocument.{self._object.Name}")

        # default values: color
        color_prop = FreeCADGui.ActiveDocument.ActiveObject.Proxy.get_default_color_property()
        if color_prop:
            FreeCADGui.doCommand(
                f"extraction.ViewObject.{color_prop} = (Gui.ActiveDocument.{vis_obj.Name}.Proxy.get_next_default_color())"
            )

        FreeCADGui.doCommand(f"App.ActiveDocument.{vis_obj.Name}.addObject(extraction)")

        self._post_dialog._recompute()
        self.repopulate()

    QtCore.Slot(object, object)  # post object, extraction data

    def addExtractionToPostObject(self, post_obj, ext_data):

        FreeCADGui.addModule(ext_data.module)
        FreeCADGui.addModule("FemGui")

        # create extraction and add it
        FreeCADGui.doCommand(
            f"extraction = {ext_data.module}.{ext_data.factory}(FreeCAD.ActiveDocument)"
        )
        FreeCADGui.doCommand(f"extraction.Source = FreeCAD.ActiveDocument.{post_obj.Name}")

        # default values for color
        color_prop = FreeCADGui.ActiveDocument.ActiveObject.Proxy.get_default_color_property()
        if color_prop:
            FreeCADGui.doCommand(
                f"extraction.ViewObject.{color_prop} = Gui.ActiveDocument.{self._object.Name}.Proxy.get_next_default_color()"
            )

        FreeCADGui.doCommand(f"App.ActiveDocument.{self._object.Name}.addObject(extraction)")

        self._post_dialog._recompute()
        self.repopulate()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

# This code handles parameter groups following the CAM workbench renaming
# (PATH -> CAM).  This code can be removed after the 1.0 release.


ParGrp = App.ParamGet("System parameter:Modules").GetGroup("Path")
if ParGrp.HasGroup("CAM"):
    pass
elif ParGrp.HasGroup("Path"):
    ParGrp.RenameGroup("Path", "CAM")

ParGrp = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Path")
parent = ParGrp.Parent()
if parent.HasGroup("CAM"):
    pass
elif parent.HasGroup("Path"):
    result = parent.RenameGroup("Path", "CAM")

# Get the Parameter Group of this module
ParGrp = App.ParamGet("System parameter:Modules").GetGroup("Path")

# Set the needed information
ParGrp.SetString("HelpIndex", "Path/Help/index.html")
ParGrp.SetString("WorkBenchName", "CAM")
ParGrp.SetString("WorkBenchModule", "PathWorkbench.py")

FreeCAD.__unit_test__ += ["TestCAMApp"]



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM glyph filter task panel for the document object"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package task_post_glyphfilter
#  \ingroup FEM
#  \brief task panel for post glyph filter object

from PySide import QtCore, QtGui

import FreeCAD
import FreeCADGui

from femguiutils import selection_widgets
from . import base_fempostpanel


class _TaskPanel(base_fempostpanel._BasePostTaskPanel):
    """
    The TaskPanel for editing properties of glyph filter
    """

    def __init__(self, vobj):
        super().__init__(vobj.Object)

        # glyph parameter widget
        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/TaskPostGlyph.ui"
        )
        self.widget.setWindowIcon(FreeCADGui.getIcon(":/icons/FEM_PostFilterGlyph.svg"))
        self.__init_widget()

        # form made from param and selection widget
        self.form = [self.widget, vobj.createDisplayTaskWidget()]

    # Setup functions
    # ###############

    def __init_widget(self):

        # set current values to ui
        self._enumPropertyToCombobox(self.obj, "Glyph", self.widget.FormComboBox)
        self._enumPropertyToCombobox(self.obj, "OrientationData", self.widget.OrientationComboBox)
        self._enumPropertyToCombobox(self.obj, "ScaleData", self.widget.ScaleComboBox)
        self._enumPropertyToCombobox(self.obj, "VectorScaleMode", self.widget.VectorModeComboBox)
        self._enumPropertyToCombobox(self.obj, "MaskMode", self.widget.MaskModeComboBox)

        self.widget.ScaleFactorBox.setValue(self.obj.ScaleFactor)
        self.__slide_min = self.obj.ScaleFactor * 0.5
        self.__slide_max = self.obj.ScaleFactor * 1.5
        self.widget.ScaleSlider.setValue(50)
        self.widget.StrideBox.setValue(self.obj.Stride)
        self.widget.MaxBox.setValue(self.obj.MaxNumber)
        self.__update_scaling_ui()
        self.__update_masking_ui()

        # connect all signals
        self.widget.FormComboBox.currentTextChanged.connect(self._form_changed)
        self.widget.OrientationComboBox.currentTextChanged.connect(self._orientation_changed)
        self.widget.ScaleComboBox.currentTextChanged.connect(self._scale_data_changed)
        self.widget.VectorModeComboBox.currentTextChanged.connect(self._scale_vector_mode_changed)
        self.widget.ScaleFactorBox.valueChanged.connect(self._scale_factor_changed)
        self.widget.ScaleSlider.valueChanged.connect(self._scale_slider_changed)
        self.widget.MaskModeComboBox.currentTextChanged.connect(self._mask_mode_changed)
        self.widget.StrideBox.valueChanged.connect(self._stride_changed)
        self.widget.MaxBox.valueChanged.connect(self._max_number_changed)

    def __update_scaling_ui(self):
        enabled = self.widget.ScaleComboBox.currentIndex() != 0
        self.widget.VectorModeComboBox.setEnabled(enabled)
        self.widget.ScaleFactorBox.setEnabled(enabled)
        self.widget.ScaleSlider.setEnabled(enabled)

    def __update_masking_ui(self):
        enabled = self.widget.MaskModeComboBox.currentIndex() != 0
        self.widget.StrideBox.setEnabled(enabled)
        self.widget.MaxBox.setEnabled(enabled)

    # callbacks and logic
    # ###################

    def _form_changed(self, value):
        self.obj.Glyph = value
        self._recompute()

    def _orientation_changed(self, value):
        self.obj.OrientationData = value
        self._recompute()

    def _scale_data_changed(self, value):
        self.obj.ScaleData = value
        self._enumPropertyToCombobox(self.obj, "VectorScaleMode", self.widget.VectorModeComboBox)
        self.__update_scaling_ui()
        self._recompute()

    def _scale_vector_mode_changed(self, value):
        self.obj.VectorScaleMode = value
        self._recompute()

    def _scale_factor_changed(self, value):

        # set slider
        self.__slide_min = value * 0.5
        self.__slide_max = value * 1.5
        slider_value = (value - self.__slide_min) / (self.__slide_max - self.__slide_min) * 100.0
        self.widget.ScaleSlider.blockSignals(True)
        self.widget.ScaleSlider.setValue(slider_value)
        self.widget.ScaleSlider.blockSignals(False)

        self.obj.ScaleFactor = value
        self._recompute()

    def _scale_slider_changed(self, value):

        # calculate value
        #                                 ( max - min )
        # factor = min + ( slider_value x ------------- )
        #                                     100
        #
        f = self.__slide_min + (value * (self.__slide_max - self.__slide_min) / 100)

        # sync factor spin box
        self.widget.ScaleFactorBox.blockSignals(True)
        self.widget.ScaleFactorBox.setValue(f)
        self.widget.ScaleFactorBox.blockSignals(False)

        # set value
        self.obj.ScaleFactor = f
        self._recompute()

    def _mask_mode_changed(self, value):
        self.obj.MaskMode = value
        self.__update_masking_ui()
        self._recompute()

    def _stride_changed(self, value):
        self.obj.Stride = value
        self._recompute()

    def _max_number_changed(self, value):
        self.obj.MaxNumber = value
        self._recompute()
